/*
 * led4x.c
 *
 *  Created on: 14.6.2011
 *      Author: emlin
 */

#include "led4x-ioctl.h"

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/errno.h>
#include <linux/device.h>
#include <linux/err.h>
#include <linux/kernel.h>

#include <asm/uaccess.h>

#include <linux/ioport.h>
#include <asm/io.h>

#include <linux/jiffies.h>
#include <linux/timer.h>

// -----------------------------------------------------------------------------
#define DEV_NAME					"led4x"
#define DEV_COUNT						1

#define PAR_PORT_BASE				0x378
#define PAR_PORT_SIZE				3
// ofset od zacatku bloku registru paralelniho portu
#define PAR_PORT_REG_DATA			(0x00)
#define PAR_PORT_REG_STAT			(0x01)
#define PAR_PORT_REG_CTRL			(0x02)

// -----------------------------------------------------------------------------
// Hodnoty signalu /SEL, /AF a /STR jsou HW invertovane
#define	LED4X_WRITE_VALUE		0x0F	// /SEL = 1, /INT = 1, /AF = 1, /STR = 1
#define LED4X_HOLD_VALUE		0x0B	// /SEL = 1, /INT = 0, /AF = 1, /STR = 1

// -----------------------------------------------------------------------------
enum
{
	ModuleExit = 0,

	Failed_ChrdevRegion,
	Failed_CdevAlloc,
	Failed_CdevAdd,
	Failed_SysClass,
	Failed_SysDev,
	Failed_ReqRegion,
	Failed_IoMap,
};

// -----------------------------------------------------------------------------
// parametry ovladace
struct led4x_t
{
	// major-minor par
	dev_t				dev_num;
	// ukazatel na strukturu cdev reprezentujici znakove zarizeni
	struct cdev *		cdev_ptr;
	// ukazatel na tridu zarizeni v adresari /sys/class
	struct class *		sys_class_ptr;
	struct device * 	sys_dev_ptr;
	// hlida pocet volani operace open()
	atomic_t			open_cnt;
	// zacatek bloku registru paralelniho portu namapovaneho jako I/O pamet
	void __iomem *		port_base;
	// casovac
	struct timer_list	blink_timer;
	// perioda casovace (milisec)
	atomic_t 			blink_timer_period;
	// promenna pro deaktivaci casovace
	// (1 = deaktivuj casovac, 0 = casovac byl aktivovan
	atomic_t 			blink_timer_stop;
	// sdilena promena - posledni hodnota zapsana do LED diod
	unsigned char 		last_value;
	// spinlock pro pristup ke sdilene promenne
	spinlock_t 			last_value_lock;
};

static struct led4x_t leds =
{
	.dev_num = MKDEV(0, 0),
	.cdev_ptr = NULL,
	.sys_class_ptr = NULL,
	.sys_dev_ptr = NULL,
	.open_cnt = ATOMIC_INIT(0),
	.port_base = NULL,
	.blink_timer_period = ATOMIC_INIT(100),
	.blink_timer_stop = ATOMIC_INIT(1),
	.last_value = 0x00,
	.last_value_lock = SPIN_LOCK_UNLOCKED,
};

// -----------------------------------------------------------------------------
static void set_leds_value(unsigned char value)
{
	iowrite8(LED4X_WRITE_VALUE, (leds.port_base + PAR_PORT_REG_CTRL));
	wmb();
	iowrite8((~(value)) & 0xFF, (leds.port_base + PAR_PORT_REG_DATA));
	wmb();
	iowrite8(LED4X_HOLD_VALUE, (leds.port_base + PAR_PORT_REG_CTRL));
}

// -----------------------------------------------------------------------------
static inline void blinktimer_on(void)
{
	int period;

	// vynulujeme si deaktivacni promenou
	atomic_set(&leds.blink_timer_stop, 0);

	// vycteme si aktualni periodu (milisecs)
	period = atomic_read(&leds.blink_timer_period);

	// nastavime si znacku, kdy ma casovac vyprset
	leds.blink_timer.expires = jiffies + msecs_to_jiffies(period);

	// spustime casovac
	add_timer(&leds.blink_timer);
}

// -----------------------------------------------------------------------------
static inline void blinktimer_off(void)
{
	atomic_set(&leds.blink_timer_stop, 1);
	del_timer(&leds.blink_timer);
}

// -----------------------------------------------------------------------------
void blinktimer_user_function(unsigned long data)
{
	unsigned char tmp_var = 0x00;

	// - zamceme pristup ke sdilene promene (je zakazano SW IRQ,
	//   tedy nefunguje docasne casovac)
	// - v kriticke sekci invertujeme hodnotu LEdek a tuto novou hodnotu
	//   si vycteme ze sdilene promenne
	spin_lock_bh(&leds.last_value_lock);

	leds.last_value = ~leds.last_value;
	tmp_var = leds.last_value;

	spin_unlock_bh(&leds.last_value_lock);

	// preneseme novou hodnotu do LEded.
	set_leds_value(tmp_var);

	// pokud nemame vypnout casovac, tak jej naplanujeme znova
	if ( !atomic_read(&leds.blink_timer_stop) )
	{
		mod_timer(&leds.blink_timer,
				  jiffies + msecs_to_jiffies(atomic_read(
						  							&leds.blink_timer_period)));
	}
}

// -----------------------------------------------------------------------------
static void cmd_enable_leds_blinking(int enable)
{
	if ( enable )
	{
		blinktimer_on();
	}
	else
	{
		blinktimer_off();
	}
}

// -----------------------------------------------------------------------------
static void cmd_set_leds_blinking_period(int period)
{
	atomic_set(&leds.blink_timer_period, period);
}

// -----------------------------------------------------------------------------
static int led4x_open(struct inode * inode_ptr, struct file * file_ptr)
{
	int result = 0, tmp_open_cnt = 0;

	// soubor zarizeni muze byt otevren jen jednim programem soucasne
	// atomicky inkrementuje promenou o 1 a vrati PUVODNI hodnotu
	tmp_open_cnt = atomic_inc_and_test(&leds.open_cnt);
	if ( tmp_open_cnt > 1 )
	{
		pr_err("led4x_open(): already opened!\n");
		atomic_dec(&leds.open_cnt);
		result = -EBUSY;
		goto End;
	}

	// modul nepodporuje systemove volani lseek()
	result = nonseekable_open(inode_ptr, file_ptr);
	if ( result < 0 )
	{
		goto End;
	}

End:
	pr_devel("led4x_open(): %d\n", result);

	return result;
}

// -----------------------------------------------------------------------------
static int led4x_release(struct inode * inode_ptr, struct file * file_ptr)
{
	// vypneme casovac (pokud nebezel nic se nestane)
	blinktimer_off();

	// povolime nove otevreni
	atomic_dec(&leds.open_cnt);

	pr_devel("led4x_release(): done\n");

	return 0;
}

// -----------------------------------------------------------------------------
// Funkce: Vycte posledni hodnotu zapsanou do LEDek.
//
static ssize_t led4x_read(struct file * file_ptr,
				   		  char __user * data_ptr, size_t data_size,
				   		  loff_t * pos_ptr)
{
	ssize_t result = 0;

	if ( data_size > 0 )
	{
		unsigned char tmp_var;

		// - zamceme pristup ke sdilene promene (je zakazano SW IRQ,
		//   tedy nefunguje docasne casovac)
		// - v kriticke sekci si vycteme hodnotu LEDed
		spin_lock_bh(&leds.last_value_lock);

		tmp_var = leds.last_value;

		spin_unlock_bh(&leds.last_value_lock);

		pr_devel("R: 0x%x\n", tmp_var);

		// zapis jej do uzivatelskeho prostoru
		// a kdyz je to ok, tak vrat velikost promenne v bytech
		result = put_user(tmp_var, data_ptr);
		if ( result == 0 )
		{
			result = sizeof(tmp_var);
		}
	}

	pr_devel("led4x_read(): %d\n", result);

	return result;
}

// -----------------------------------------------------------------------------
// Funkce: Zapise novou hodnotu do LEDek.
//
static ssize_t led4x_write(struct file * file_ptr,
				    	   const char __user * data_ptr, size_t data_size,
				    	   loff_t * posPtr)
{
	ssize_t result = 0;

	if ( data_size > 0 )
	{
		unsigned char tmp_var;

		// zkopirujeme novou hodnotu LEDed z uzivatelskeho prostoru
		result = get_user(tmp_var, data_ptr);
		if ( result < 0 )
		{
			goto End;
		}

		result = sizeof(tmp_var);

		printk(KERN_DEBUG "W: 0x%x\n", tmp_var);

		// - zamceme pristup ke sdilene promene (je zakazano SW IRQ,
		//   tedy nefunguje docasne casovac)
		// - v kriticke sekci zapiseme novou hodnotu do promenne
		spin_lock_bh(&leds.last_value_lock);

		leds.last_value = tmp_var;

		spin_unlock_bh(&leds.last_value_lock);

		// preneseme novou hodnotu do LEDek
		set_leds_value(tmp_var);
	}

End:
	pr_devel("led4x_write(): %d\n", result);

	return result;
}

// -----------------------------------------------------------------------------
static int led4x_ioctl(struct inode * inode_ptr, struct file * file_ptr,
				       unsigned int cmd, unsigned long arg)
{
	int result = 0, tmp_var = 0;

	// kontrola kodu prikazu
	if ( (_IOC_TYPE(cmd) != LED4X_MAGIC_NO) || (_IOC_NR(cmd) > LED4X_MAX_MCD_NO) )
	{
		result = -ENOTTY;
		goto End;
	}

	// kontrola argumentu (platnost ukazatele a smeru pristupu)
	// Pozn.: smery v access_ok() jsou presne opacne nez u ioctl
	if ( _IOC_DIR(cmd) & _IOC_READ )
	{
		result = access_ok(VERIFY_WRITE, (void __user *) arg, _IOC_SIZE(cmd));
	}
	else if ( _IOC_DIR(cmd) & _IOC_WRITE )
	{
		result = access_ok(VERIFY_READ, (void __user *) arg, _IOC_SIZE(cmd));
	}
	else
	{ ; /*empty*/ }

	if ( result < 0 )
	{
		goto End;
	}

	// dekodujeme prikaz a pak jej provedeme
	switch( cmd )
	{
		case LED4X_EN_BLINKING:
		{
			result = __get_user(tmp_var, (int *) arg);
			if ( result < 0 )
			{
				goto End;
			}

			cmd_enable_leds_blinking(tmp_var);
		}
		break;

		case LED4X_SET_PERIOD:
		{
			result = __get_user(tmp_var, (int *) arg);
			if ( result < 0 )
			{
				goto End;
			}

			cmd_set_leds_blinking_period(tmp_var);
		}
		break;

		default:
			result = -ENOTTY;
			break;
	}

End:
	pr_devel("led4x_ioctl(): %d\n", result);

	return result;
}

// -----------------------------------------------------------------------------
// virtualni tabulka - tabulka operaci podporovanych ovladacem
static struct file_operations led4x_file_ops =
{
	.owner = THIS_MODULE,
	.open = &led4x_open,
	.release = &led4x_release,
	.read = &led4x_read,
	.write = &led4x_write,
	.ioctl = &led4x_ioctl,
	// modul nepodporuje metodu llseek
	.llseek = &no_llseek,
};

// -----------------------------------------------------------------------------
static void cleanup(int error)
{
	switch ( error )
	{
		// kompletni uklid pri exitu
		case ModuleExit:
			// zrusime adresar modulu v /sys
			device_destroy(leds.sys_class_ptr, leds.dev_num);

		case Failed_SysDev:
			// zrusime tridu
			class_destroy(leds.sys_class_ptr);

		case Failed_SysClass:
			// vyjmeme modul z jadra
			// pozn.: cdev_del() se vola jen po uspesnem cdev_add()
			cdev_del(leds.cdev_ptr);

		case Failed_CdevAdd:
			// vypneme casovac (pokud nebezel nic se nestane)
			blinktimer_off();
			// zrusime mapovani portu
			ioport_unmap(leds.port_base);

		case Failed_IoMap:
			// zrusime rezervaci I/O portu
			release_region(PAR_PORT_BASE, PAR_PORT_SIZE);

		case Failed_ReqRegion:
			// pozn.: neexistuje opak fce cdev_alloc(), uvolneni pameti resi
			// jadro samo
		case Failed_CdevAlloc:
			// uvolnime nami zabrany major-minor par
			unregister_chrdev_region(leds.dev_num, DEV_COUNT);

		case Failed_ChrdevRegion:
			// jeste nic se nealokovalo, takze neni co vracet
			break;

		default:
			break;
	}
}

// -----------------------------------------------------------------------------
static int __init led4x_init(void)
{
	int result = 0;
	struct resource * region_ptr;

	// registrace k major-minor paru
	result = alloc_chrdev_region(&leds.dev_num, MINOR(leds.dev_num),
								 DEV_COUNT, DEV_NAME);
	if ( result < 0 )
	{
		cleanup(Failed_ChrdevRegion);
		goto End;
	}

	// alokace pameti pro strukturu cdev
	leds.cdev_ptr = cdev_alloc();
	if ( leds.cdev_ptr == NULL )
	{
		result = -ENOMEM;
		cleanup(Failed_CdevAlloc);
		goto End;
	}

	// nastaveni vlastnika a seznamu souborovych operaci
	leds.cdev_ptr->owner = THIS_MODULE;
	leds.cdev_ptr->ops = &led4x_file_ops;

	// rezervujeme si paralelni port
	region_ptr = request_region(PAR_PORT_BASE, PAR_PORT_SIZE, DEV_NAME);
	if ( region_ptr == NULL )
	{
		pr_err("led4x_init(): parallel port already in use!\n");
		cleanup(Failed_ReqRegion);
		result = -EBUSY;
		goto End;
	}

	// premapujem port jako I/O pamet
	leds.port_base = ioport_map(PAR_PORT_BASE, PAR_PORT_SIZE);
	if ( leds.port_base == NULL )
	{
		cleanup(Failed_IoMap);
		result = -ENOMEM;
		goto End;
	}

	// inicializace casovace (zbytek se nastavuje pri spusteni)
	leds.blink_timer.function = &blinktimer_user_function;
	leds.blink_timer.data = (unsigned long) NULL;
	init_timer(&leds.blink_timer);

	// vypnuti LEDek jako vychozi stav
	set_leds_value(0x00);

	//  --- pridani ovladace do jadra ---
	result = cdev_add(leds.cdev_ptr, leds.dev_num, DEV_COUNT);
	if ( result < 0 )
	{
		cleanup(Failed_CdevAdd);
		goto End;
	}

	// vytvorime si tridu v /sys/class
	leds.sys_class_ptr = class_create(THIS_MODULE, DEV_NAME);
	if ( IS_ERR(leds.sys_class_ptr) )
	{
		cleanup(Failed_SysClass);
		result = PTR_ERR(leds.sys_class_ptr);
		goto End;
	}

	// pridame nas modul do /sys/class/led4x
	leds.sys_dev_ptr = device_create(leds.sys_class_ptr,
									 NULL, leds.dev_num, NULL, DEV_NAME);
	if ( IS_ERR(leds.sys_dev_ptr) )
	{
		cleanup(Failed_SysDev);
		result = PTR_ERR(leds.sys_dev_ptr);
		goto End;
	}

End:
	pr_devel("led4x_init(): %d\n", result);

	return result;
}

// -----------------------------------------------------------------------------
static void __exit led4x_exit(void)
{
	cleanup(ModuleExit);

	pr_devel("led4x_exit(): done\n");
}

// -----------------------------------------------------------------------------
module_init(led4x_init);
module_exit(led4x_exit);

MODULE_AUTHOR("mBoss <marek.laznicka@elvoris.cz>");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Simple four LEDs driver");
MODULE_VERSION("0.0.1");
